#include "mysocket.h"

#include <cstring>
#include <unistd.h>
#include <sys/ioctl.h>

#include "mymisc.h"

namespace mysocket_ns
{

// Sulle diskless powerpc c'e' un baco del kernel (risolto poi nel 2.6.28)
// che rende non bloccanti chiamate recv() che contengono contemporaneamente
// i flags MSG_PEEK e MSG_WAITALL (in questo caso infatti il flag MSG_WAITALL
// viene ignorato).

void MySocket::set_timeout(unsigned int timeout_ms)
{
	timeout.tv_sec = timeout_ms / 1000;
	timeout.tv_usec = (timeout_ms - (timeout.tv_sec * 1000)) * 1000;
}

void MySocket::set_ltimeout(unsigned int ltimeout_sec)
{
	if (ltimeout_sec > 0) {
		struct linger so_linger;
		so_linger.l_onoff = 1;
		so_linger.l_linger = ltimeout_sec;
		setsockopt(socketfd, SOL_SOCKET, SO_LINGER, &so_linger, sizeof so_linger);
	}
}

void MySocket::flush(void)
{
	int len = length();
	unsigned char *buff = (unsigned char*)malloc(len);
	recv(socketfd, buff, len, MSG_DONTWAIT);
	free(buff);

	TANGO_DEBUG << "Flushed " << len << " bytes";
}

int MySocket::length(void)
{
	int to_read;

	if (ioctl(socketfd, FIONREAD, &to_read) == -1)
		TANGO_ERROR << "The ioclt() failed!";

	return to_read;
}

void MySocket::dump(const unsigned char *databuff, const int datalen)
{
	stringstream hex;
	int len;
	unsigned char *buff;
	char tmp[3];

	if (databuff != NULL) {
		// I'll dump buffer provided
		len = datalen;
		buff = (unsigned char*)malloc(len + 1);
		memcpy(buff, databuff, len);
	} else {
		// I'll dump the system buffer
		len = length();
		buff = (unsigned char*)malloc(len + 1);
		recv(socketfd, buff, len, MSG_PEEK | MSG_DONTWAIT);
	}
	buff[len] = '\0';

	for (int i=0; i< len; i++) {
		sprintf(tmp, "%02x", buff[i]);
		hex << tmp;
	}

	TANGO_DEBUG << "Buffer length: " << len  << " Contents in hex: " << hex.str() << " Contents in ASCII: " << buff;

	free(buff);
}

bool MySocket::wait_data(struct timeval *tv)
{
	FD_ZERO(&readfds);
	FD_SET(socketfd, &readfds);
	FD_SET(pipefd[0], &readfds);

	if (interrupted)
		return false;

	if (select(maxfd + 1, &readfds, NULL, NULL, tv) < 1)
		return false;

	if (! FD_ISSET(socketfd, &readfds))
		return false;

	return true;
}

int MySocket::read_for(std::stringstream& strbuffer, const int numchars)
{
	int ret;
	unsigned char *buff = (unsigned char *)malloc(numchars);

	ret = read_for(buff, numchars);
	if (ret > 0)
		strbuffer.rdbuf()->sputn((char *)buff, ret); // Copy the characters requested into a stringstream

	free(buff);
	return ret;
}

int MySocket::read_for(unsigned char* buff, const int numbytes)
{
	struct timeval tv = timeout;
	int total = 0; // how many bytes we've received
	int bytesleft = numbytes; // how many bytes missing
	int n;

	while (total < numbytes) {
		// Wait for new data
		if (! wait_data(&tv))
			return total;

		// Get data from buffer
		n = recv(socketfd, buff+total, bytesleft, MSG_DONTWAIT);
		if (n == -1) { return -1; }
		total += n;
		bytesleft -= n;
	}

	return total;
}

bool MySocket::read_until(std::stringstream& strbuffer, const char lastchar)
{
	struct timeval tv = timeout;
	char tmp;
	char *buff = NULL;
	int len = 0;
	int pos = 0;

	while (1) {
		// Wait for new data
		if (! wait_data(&tv)) {
			strbuffer.rdbuf()->sputn(buff, pos); // Copy the characters requested into a stringstream
			free(buff);
			return false;
		}

		// Copy available data into a local buffer
		len = length();
		buff = (char*)realloc(buff, pos+len);
		recv(socketfd, buff+pos, len, MSG_PEEK | MSG_DONTWAIT);

		// Loking for the terminator
		for (int i=0; i<len; i++) {
			if (buff[pos+i] == lastchar) {
				recv(socketfd, &tmp, sizeof(tmp), MSG_DONTWAIT); // Remove terminator from socket
				strbuffer.rdbuf()->sputn(buff, pos+i); // Copy the characters requested into a stringstream
				free(buff);
				return true;
			}
			recv(socketfd, &tmp, sizeof(tmp), MSG_DONTWAIT); // Remove a byte from socket
		}

		// Move available data into a local buffer
		recv(socketfd, buff+pos, len, MSG_DONTWAIT);

		pos += len;
	}
}

/*
int MySocket::send_one_at_a_time(const std::stringstream& strbuffer, const unsigned int delay_between_chars)
{
	int total = 0;        // how many bytes we've sent
	int len = strbuffer.str().size();
	int n;

	if (nagle)
		return false;

	while(total < len) {
		n = send(socketfd, strbuffer.str().c_str()+total, 1, MSG_NOSIGNAL);
		if (n == -1) { return n; }
		total += n;
		usleep(delay_between_chars);
	}

	return total;
}
*/

int MySocket::send_all(const std::stringstream& strbuffer)
{
	return send_all((unsigned char *)strbuffer.str().c_str(), strbuffer.str().size());
}

int MySocket::send_all(const unsigned char *buff, const int numbytes)
{
	int total = 0;        // how many bytes we've sent
	int bytesleft = numbytes; // how many we have left to send
	int n;

	while(total < numbytes) {
		n = send(socketfd, buff+total, bytesleft, MSG_NOSIGNAL);
		if (n == -1) { return n; }
		total += n;
		bytesleft -= n;
	}

	return total;
}

MySocket::MySocket(class Tango::DeviceImpl *td, const std::string host_to_connect_to,
		const unsigned short port_to_connect_to, const int timeout_ms, bool nstatus)
	: Tango::LogAdapter(td), nagle(nstatus)
{
	tango_device = td;
	socketfd = -1;
	hostname = host_to_connect_to;
	port = port_to_connect_to;
	set_timeout(timeout_ms);
}

bool MySocket::connect(void)
{
	char ipstr[INET6_ADDRSTRLEN];

	struct sockaddr_in sa;
	memset(&sa, 0, sizeof(sa));

	struct addrinfo hints;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	struct addrinfo *res, *p;

	if( inet_pton(AF_INET, hostname.c_str(), &(sa.sin_addr))>0 )
	{
		sa.sin_family = AF_INET;
		sa.sin_port = htons(port);
	}
	else if( getaddrinfo(hostname.c_str(), NULL, &hints, &res)==0 )
	{
		for(p=res; p!=NULL; p=p->ai_next)
		{
			if(p->ai_family == AF_INET)
			{
				inet_ntop(p->ai_family, &(((struct sockaddr_in *)p->ai_addr)->sin_addr), ipstr, sizeof ipstr);
				sa.sin_addr = ((struct sockaddr_in *)p->ai_addr)->sin_addr;
				sa.sin_family = ((struct sockaddr_in *)p->ai_addr)->sin_family;
				sa.sin_port = htons(port);
			}
		}        
		freeaddrinfo(res);
	}
	else
		return false;

	socketfd = socket(PF_INET, SOCK_STREAM, 0);

	/* Disable or enable the Nagle algorithm */
	int flag = nagle ? 0 : 1;
	if (! setsockopt(socketfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag)) < 0) {
		nagle = nagle ? false : true;
	}

	/* Set the default linger timeout */
	set_ltimeout(1);

	/* Finally connect */
	if( ::connect(socketfd, (struct sockaddr*)&sa, sizeof(struct sockaddr)) == -1 )
		return false;

	if (pipe(pipefd) < 0)
		return false;

	maxfd = pipefd[0] > socketfd? pipefd[0] : socketfd;

	interrupted = false;

	return true;
}

MySocket::~MySocket()
{
	if (is_connected())
		disconnect();
}

void MySocket::interrupt(void)
{
	interrupted = true;
	int d = 1;
	ssize_t ret;
	ret = write(pipefd[1], &d, sizeof(d));
	(void)ret;
}

void MySocket::disconnect(void)
{
	close(socketfd);
	close(pipefd[0]);
	close(pipefd[1]);
	socketfd = -1;
}

bool MySocket::is_connected(void)
{
	unsigned char buff;
	int ret;

	if (socketfd <= 0)
		return false;

	ret = recv(socketfd, &buff, 1, MSG_PEEK | MSG_DONTWAIT);

	if (ret != 0)
		return true;
	return false;
}

bool MySocket::check_and_reconnect(stringstream &strerr)
{
	if (is_connected())
		return true;

	if (connect())
		return true;

	strerr << strerror(errno);

	disconnect();

	return false;
}

} // namespace mysocket_ns
